# -*- coding: utf-8 -*-

"""Provides a ``includeme()`` pyramid configuration entry point."""

import logging
logger = logging.getLogger(__name__)

import os

from pyramid import authorization
from pyramid import security
from pyramid.settings import asbool

import notification as n

from . import auth

DEFAULTS = {
    'notification.api_key': os.environ.get('PYRAMID_NOTIFICATION_API_KEY'),
}


class IncludeMe(object):
    """Set up the state change event subscription system and provide an
      ``add_engine_subscriber`` directive.
    """

    def __init__(self, **kwargs):
        self.add_notification = kwargs.get('add_notification', n.add_notification)
        self.add_roles_mapping = kwargs.get('add_roles_mapping', n.add_roles_mapping)
        self.get_roles_mapping = kwargs.get('get_roles_mapping', n.get_roles_mapping)

    def __call__(self, config):
        """Handle `/events` requests and provide subscription directive."""

        # Dispatch the notifications.
        config.add_request_method(n.dispatch_notifications, 'dispatch_notifications', reify=True)

        # Adds a notification to the resource.
        config.add_directive('add_notification', self.add_notification)
        config.registry.roles_mapping = {}

        # Adds / gets role mapping.
        config.add_directive('add_roles_mapping', self.add_roles_mapping)
        config.add_directive('get_roles_mapping', self.get_roles_mapping)

        # Operator user to receive admin related emails.
        config.add_request_method(n.get_operator_user, 'operator_user', reify=True)

        # Email sender.
        config.include('pyramid_postmark')

        # Expose webhook views to notifications such as single / batch emails / sms's.
        config.add_route('notification_single', '/notifications/single')
        config.add_view(n.notification_single_view, renderer='json',
                request_method='POST', route_name='notification_single')

        config.add_route('notification_batch', '/notifications/batch')
        config.add_view(n.notification_batch_view, renderer='json',
                request_method='POST', route_name='notification_batch')


includeme = IncludeMe().__call__



# -*- coding: utf-8 -*-

"""Provides a base ``ACLContainer`` class that wraps any context in a
  generic API key aware ``ACLWrapper`` and a
"""

__all__ = [
    'APIKeyAuthenticationPolicy',
    'APIKeyAuthorizationPolicy',
]

import logging
logger = logging.getLogger(__name__)

import re
import zope.interface as zi

from pyramid import authentication
from pyramid import interfaces

VALID_API_KEY = re.compile(r'^\w{40}$')

@zi.implementer(interfaces.IAuthenticationPolicy)
class APIKeyAuthenticationPolicy(authentication.CallbackAuthenticationPolicy):
    """A Pyramid authentication policy which obtains credential data from the
      ``request.headers['api_key']``.
    """

    def __init__(self, header_keys, **kwargs):
        if isinstance(header_keys, basestring):
            header_keys = [header_keys]
        self.header_keys = header_keys
        self.valid_key = kwargs.get('valid_key', VALID_API_KEY)

    def unauthenticated_userid(self, request):
        """The ``api_key`` value found within the ``request.headers``."""

        api_key = None
        for key in self.header_keys:
            value = request.headers.get(key, None)
            if value is not None:
                api_key = value
                break
        if api_key and self.valid_key.match(api_key):
            return api_key.decode('utf8')

    def remember(self, request, principal, **kw):
        """A no-op. There's no way to remember the user."""

        return []

    def forget(self, request):
        """A no-op. There's no user to forget."""

        return []

@zi.implementer(interfaces.IAuthorizationPolicy)
class APIKeyAuthorizationPolicy(object):
    """Global authorization policy that ignores the context and just checks
      whether the target api key is in the principals list.
    """

    def __init__(self, api_key):
        self.api_key = api_key

    def permits(self, context, principals, permission):
        return self.api_key in principals

    def principals_allowed_by_permission(self, context, permission):
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""Utility functions."""

import logging
logger = logging.getLogger(__name__)

def extract_from(request):
    settings = request.registry.settings
    site_email = settings.get('site.email')
    site_title = settings.get('site.title')
    return u'{0} <{1}>'.format(site_title, site_email)



# -*- coding: utf-8 -*-

from sqlalchemy import create_engine
from pyramid_basemodel import bind_engine, save, Session
from pyramid_notification import DEFAULTS

from . import util
from . import repo
from . import orm

import os
import datetime
import json
import requests
import transaction

AVAILABLE_CHANNELS = ['sms', 'email']

env = os.environ
NOTIFICATION_SINGLE_ENDPOINT = env.get('NOTIFICATION_SINGLE_ENDPOINT', None)


def post_notification_dispatch(dispatch):

    headers = {}
    for item in c.ENGINE_API_KEY_NAMES:
        key = '{0}'.format(item)
        headers[key] = DEFAULTS['notification.api_key']

    _ = requests.post(
                    NOTIFICATION_SINGLE_ENDPOINT,
                    headers=headers,
                    data=json.dumps(
                        {'notification_dispatch_id': dispatch.id}))

def dispatch_user_notifications(user, user_notifications):
    """ 4. for each channel loop and either write out a single or a batch dispatch task with the
        NotificationDispatcher ids e.g: /dispatch_email, /dispatch_sms and etc.
    """

    for ch in AVAILABLE_CHANNELS:
        # XXX check for preferences e.g: and user.channel == ch
        to_dispatch = [d for d in user_notifications if d.category == ch]
        for dispatch in to_dispatch:
            post_notification_dispatch(dispatch)
        else:
            print 'nothing here', to_dispatch
    Session.flush()


def run():
    # Bind to the database.
    engine = create_engine(os.environ['DATABASE_URL'])
    bind_engine(engine, should_create=False)

    # Prepare.
    notification_cls = orm.Notification
    notification_dispatch_cls = orm.NotificationDispatch
    notification_preference_factory = repo.NotificationPreferencesFactory()
    now = datetime.datetime.now()

    # Run the algorithm.
    with transaction.manager:
        # 1. ignore all the notifications from the Notification table that have read field set.
        unread_notifications = notification_dispatch_cls.query.join(notification_cls).filter(notification_cls.read == None)

        # 2. get all of the non duplicated user ids who are due to dispatch and have not been sent.
        due_to_dispatch = unread_notifications.filter(notification_dispatch_cls.due <= now).filter(notification_dispatch_cls.sent == None)
        user_ids_to_dispatch = set()
        for dispatch in due_to_dispatch.all():
            user_ids_to_dispatch.add(dispatch.notification.user_id)

        # 3. for each user id get all of the notifications grouped by channel
        for user_id in user_ids_to_dispatch:
            # Build the NotificationPreference object so we can get the preferences.
            user = orm.NotificationPreference.query.filter_by(user_id=user_id).all()[-1]
            # If we don't have a notification preference object, we just create it on the fly.
            if user is None:
                user = notification_preference_factory(user_id)
            user_notifications = due_to_dispatch.filter(notification_cls.user_id == user_id).all()
            dispatch_user_notifications(user, user_notifications)


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

"""Provides an dummy model for ftesting."""

import transaction as tx
import zope.interface as zi

import pyramid_basemodel as bm

from pyramid_torque_engine import orm

class IModel(zi.Interface):
    pass

class IContainer(zi.Interface):
    pass

class IFoo(IModel):
    pass

class IFooContainer(IContainer):
    pass

class IBar(IModel):
    pass

class IBarContainer(IContainer):
    pass

@zi.implementer(IModel)
class Model(bm.Base, bm.BaseMixin, orm.WorkStatusMixin):
    __tablename__ = 'models'

    def __json__(self, request=None):
        return {}

@zi.implementer(IFoo)
class Foo(bm.Base, bm.BaseMixin, orm.WorkStatusMixin):
    __tablename__ = 'foos'

    def __json__(self, request=None):
        return {}

@zi.implementer(IBar)
class Bar(bm.Base, bm.BaseMixin, orm.WorkStatusMixin):
    __tablename__ = 'bars'

    def __json__(self, request=None):
        return {}

def factory(cls=Model, initial_state=orm.DEFAULT_STATE, **kwargs):
    with tx.manager:
        instance = cls(**kwargs)
        instance.set_work_status(initial_state)
        bm.Session.add(instance)
        bm.Session.flush()
        instance_id = instance.id
    return cls.query.get(instance_id)



# -*- coding: utf-8 -*-

"""Provides sync for podio."""

__all__ = [
    'add_notification',
    'AddNotification'
]

import logging
logger = logging.getLogger(__name__)

from pyramid_torque_engine import unpack
from pyramid_torque_engine import operations as ops

from . import repo
from . import util
from pyramid import path

from pyramid_simpleauth.model import get_existing_user

import colander
import datetime
import pyramid_basemodel as bm
import requests
import json
import os


def send_from_notification_dispatch(request, notification_dispatch_id):
    """Boilerplate to extract information from the notification
    dispatch and send an email.
    Please note that no verification if it should
    be sent is made prior to sending.
    """

    lookup = repo.LookupNotificationDispatch()
    dotted_name_resolver = path.DottedNameResolver()

    notification_dispatch = lookup(notification_dispatch_id)
    if not notification_dispatch:
        return False

    # Extract information from the notification dispatch.
    spec = notification_dispatch.single_spec
    send_to = notification_dispatch.address
    view = dotted_name_resolver.resolve(notification_dispatch.view)
    event = notification_dispatch.notification.event
    bcc = notification_dispatch.bcc
    context = event.parent
    channel = notification_dispatch.category

    # Get the template vars.
    tmpl_vars = view(request, context, send_to, event, event.action)

    # Set some defaults for the template vars.
    tmpl_vars.setdefault('data', context)
    tmpl_vars.setdefault('to', send_to)
    tmpl_vars.setdefault('from', util.extract_from(request))
    tmpl_vars.setdefault('state_or_action', event.action)
    tmpl_vars.setdefault('event', event)
    tmpl_vars.setdefault('subject', '{0} {1}'.format(event.target, event.action))
    # Check if we should add bcc.
    if bcc:
        tmpl_vars.setdefault('bcc', bcc)

    # Extract form tmpl_vars and remove.
    subject = tmpl_vars['subject']
    to_ = tmpl_vars['to']
    from_ = tmpl_vars['from']
    del tmpl_vars['subject']
    del tmpl_vars['to']
    del tmpl_vars['from']

    # Send emails / sms.
    if channel == 'email':
        email = request.render_email(
                from_,
                to_,
                subject,
                spec,
                tmpl_vars,
                **tmpl_vars)
        request.send_email(email)
    elif channel == 'sms':
        pass
    else:
        raise Exception('Unknown channel to send the notification')

    # Set the sent info in our db.
    notification_dispatch.sent = datetime.datetime.now()
    bm.save(notification_dispatch)

    return True


def notification_single_view(request):
    """View to handle a single notification dispatch"""

    class SingleNotificationSchema(colander.Schema):
        notification_dispatch_id = colander.SchemaNode(
            colander.Integer(),
        )

    schema = SingleNotificationSchema()

    # Decode JSON.
    try:
        json = request.json
    except ValueError as err:
        request.response.status_int = 400
        return {'JSON error': str(err)}

    # Validate.
    try:
        appstruct = schema.deserialize(json)
    except colander.Invalid as err:
        request.response.status_int = 400
        return {'error': err.asdict()}

    # Get data out of JSON.
    notification_dispatch_id = appstruct['notification_dispatch_id']

    # Send the email.
    r = send_from_notification_dispatch(request, notification_dispatch_id)
    if not r:
        request.response.status_int = 404
        return {'error': u'Notification dispatch not Found.'}

    # Return 200.
    return {'dispatched': 'ok'}


def notification_batch_view(request):
    """View to handle a batch email notification dispatch"""
    pass


class AddNotification(object):
    """Standard boilerplate to add a notification."""

    def __init__(self, iface, role, dispatch_mapping, delay=None, bcc=None):
        """"""

        self.dispatch_mapping = dispatch_mapping
        self.notification_factory = repo.NotificationFactory
        self.role = role
        self.delay = delay
        self.bcc = bcc
        self.iface = iface

    def __call__(self, request, context, event, op, **kwargs):
        """"""

        # Unpack.
        dispatch_mapping = self.dispatch_mapping
        notification_factory = self.notification_factory(request)
        delay = self.delay
        bcc = self.bcc
        iface = self.iface
        role = self.role

        # Prepare.
        notifications = []

        # get relevant information.
        interested_users_func = get_roles_mapping(request, iface)
        interested_users = interested_users_func(request, context)
        for user in interested_users[role]:
            # Just user is a shorthand for context.user.
            if user == 'user':
                user = context.user
            # create the notifications.
            notification = notification_factory(event, user, dispatch_mapping, delay, bcc)
            notifications.append(notification)

        # Tries to optimistically send the notification.
        dispatch_notifications(request, notifications)


def add_notification(config,
                     iface,
                     state_or_action_changes,
                     role,
                     dispatch_mapping,
                     delay=None,
                     bcc=None):

    # Unpack.
    _, o, _, s = unpack.constants()
    _, on, _ = unpack.directives(config)

    o.register(
        'CREATE_NOTIFICATION',
    )

    create_notification_in_db = AddNotification(iface, role, dispatch_mapping, delay, bcc)
    on(iface, state_or_action_changes, o.CREATE_NOTIFICATION, create_notification_in_db)


def add_roles_mapping(config, iface, mapping):
    """Adds a roles mapping to the resource."""

    # Unpack.
    registry = config.registry

    # Noop if we've done this already.
    roles_mapping = registry.roles_mapping
    if iface in roles_mapping:
        logger.warn('Notification: roles_mapping alreaady configured for {}'.format(iface))
        return

    # Register the role mapping.
    roles_mapping[iface] = mapping


def get_roles_mapping(request, iface):
    """Gets the role mapping for the resource."""

    # Unpack.
    registry = request.registry
    roles_mapping = registry.roles_mapping

    return roles_mapping.get(iface, None)


def get_operator_user(request, registry=None):
    """We have a special user in our db representing the operator user. Here
      We look them up by username, constructed from the client server name.
      The operator should be the one to receive e-mails that target
      the website / administration.
    """

    if registry == None:
        # Unpack.
        settings = request.registry.settings
    else:
        settings = registry.settings

    # Get the user, which depends on the server.
    server = os.environ.get('INI_site__title', '')
    if server.lower() == 'opendesk':
        username = u'opendesk_operator'
    elif server.lower() == 'fabhub':
        username = u'fabhub_operator'
    else:
        raise Exception('Operator user not configured.')

    return get_existing_user(username=username)


def dispatch_notifications(request, notifications):
    """Dispatches a notification directly without waiting for the
    background process."""

    lookup = repo.LookupNotificationDispatch()
    now = datetime.datetime.now()

    # Loop through the notifications and check if we should send them.
    for notification in notifications:
        # Get our create the user preferences.
        preference = repo.get_or_create_notification_preferences(notification.user)
        # Check if its due to dispatch, if so, dispatch.
        for dispatch in lookup.by_notification_id(notification.id):
            if dispatch.due <= now:
                send_from_notification_dispatch(request, dispatch.id)




# -*- coding: utf-8 -*-

"""This package contains functional tests that use webtest.TestApp to simulate
  requests to a minimal WSGI app that uses the work engine.

  This requires the WSGI app to be setup, which requires the database to be
  setup. This takes time (largely because of the tables and relations we inherit
  from fabhub). So... in order to make the tests run reasonablly fast we:

  * create a fresh db once per test run
  * create the necessary db tables in the test setup_class method
  * run tests within a setup:begin teardown:rollback transaction
"""

import os

from sqlalchemy import engine_from_config

from pyramid_basemodel import Base
from pyramid_basemodel import Session

from . import settings

# Only run this setup once, even if running tests in parallel.
_multiprocess_shared_ = True

def setup():
    """Drop and recreate a vanilla db."""

    test_settings = settings.TEST_SETTINGS
    db_url = test_settings['sqlalchemy.url']
    db_name = db_url.split('/')[-1]
    os.system('dropdb {0}'.format(db_name))
    os.system('createdb -T template0 -E UTF8 {0}'.format(db_name))

    engine = engine_from_config(test_settings, prefix='sqlalchemy.')
    Base.metadata.create_all(engine)



# -*- coding: utf-8 -*-

"""Test settings, overridable by env var."""

__all__ = [
    'TEST_SETTINGS',
]

import os

TEST_SETTINGS = {
    'mode': 'testing',
    'basemodel.should_bind_engine': False,
    'sqlalchemy.url': os.environ.get(
        'DATABASE_URL',
        'postgresql:///torque_engine_test'
    ),
    'handle_exceptions': os.environ.get(
        'HANDLE_EXCEPTIONS',
        False
    ),
    'torque.enable_ftesting_dispatch': True,
    'engine.url': 'http://localhost',
}



# -*- coding: utf-8 -*-

"""High level integration / functional tests for the notifications."""

import logging
logger = logging.getLogger(__name__)

import json
import fysom
import transaction
import mock

import pyramid_basemodel as bm

from pyramid import config as pyramid_config

from pyramid_torque_engine import constants
from pyramid_torque_engine import operations as ops
from pyramid_torque_engine import unpack
from pyramid_torque_engine import repo
a, o, r, s = unpack.constants()

from . import boilerplate
from . import model


class TestNotifications(boilerplate.AppTestCase):
    """"""

    @classmethod
    def includeme(cls, config):
        """Setup the test configuration."""

        # Unpack.
        allow, on, after = unpack.directives(config)

        # Traverse.
        config.add_engine_resource(model.Model, model.IContainer)

        # Declare constants.
        s.register(
            'CREATED',
            'STARTED',
            'COMPLETED',
            'ABSOLUTELY_COMPLETED',
            'CANCELLED',
            'TRANSMOGRIFIED',
        )
        a.register(
            'START',
            'COMPLETE',
            'CANCEL',
            'POKE',
            'TRANSMOGRIFY',
        )

        # Get a handle on the model interface.
        IModel = model.IModel

        # Declare actions.
        allow(IModel, a.START, (s.CREATED), s.STARTED)
        allow(IModel, a.COMPLETE, (s.STARTED), s.COMPLETED)
        allow(IModel, a.COMPLETE, (s.COMPLETED), s.ABSOLUTELY_COMPLETED)
        allow(IModel, a.COMPLETE, (s.ABSOLUTELY_COMPLETED), Ellipsis)
        allow(IModel, a.CANCEL, (s.CREATED, s.STARTED), s.CANCELLED)
        allow(IModel, a.POKE, '*', Ellipsis)
        allow(IModel, a.TRANSMOGRIFY, '*', s.TRANSMOGRIFIED)

    def test_notification_factory(self):
        """Test the notification factory."""

        factory = repo.NotificationFactory(mock.Mock())

        # Create an eventand get it back.
        context = model.factory()
        event_id = boilerplate.createEvent(context)
        event = repo.LookupActivityEvent()(event_id)

        with transaction.manager:
            user = boilerplate.createUser()
            bm.Session.add(user)
            notification = factory(event, user, {})

            # It exists!
            self.assertTrue(notification)

        with transaction.manager:
            # The user didn't had preference before, he should have
            # default settings as it was lazily created on the factory.
            bm.Session.add(user)
            notification_preference = user.notification_preference
            self.assertIsNone(notification_preference.frequency)
            self.assertEqual(notification_preference.channel, 'email')



# -*- coding: utf-8 -*-

"""Create and lookup notifications."""

__all__ = [
    'NotificationFactory',
    'LookupNotification',
    'LookupNotificationDispatch',
    'NotificationPreferencesFactory',
    'get_or_create_notification_preferences',
]

import logging
logger = logging.getLogger(__name__)

import json
import pyramid_basemodel as bm

from . import orm
from . import util

import datetime
from dateutil.relativedelta import relativedelta


class NotificationFactory(object):
    """Boilerplate to create and save ``Notification``s."""

    def __init__(self, request, **kwargs):
        self.request = request
        self.notification_cls = kwargs.get('notification_cls', orm.Notification)
        self.notification_dispatch_cls = kwargs.get('notification_dispatch_cls',
                orm.NotificationDispatch)
        self.notification_preference_factory = kwargs.get('notification_preference_factory',
                NotificationPreferencesFactory())
        self.session = kwargs.get('session', bm.Session)

    def __call__(self, event, user, dispatch_mapping, delay=None, bcc=None):
        """Create and store a notification and a notification dispatch."""

        # Unpack.
        session = self.session
        request = self.request

        # Create notification.
        notification = self.notification_cls(user=user, event=event)
        session.add(notification)
        due = datetime.datetime.now()
        email = user.best_email.address

        # Get or create user preferences.
        preference = get_or_create_notification_preferences(user)
        timeframe = preference.frequency

        # If daily normalise to 20h of each day.
        if timeframe == 'daily':
            if due.hour > 20:
                due = datetime.datetime(due.year, due.month, due.day + 1, 20)
            else:
                due = datetime.datetime(due.year, due.month, due.day, 20)

        # If hourly normalise to the next hour.
        elif timeframe == 'hourly':
            due = datetime.datetime(due.year, due.month, due.day, due.hour + 1, 0)

        # Check if there's a delay in minutes add to it.
        if delay:
            delay = relativedelta(minutes=delay)
            due = due + delay
        if bcc:
            if bcc is True or bcc == '':
                bcc = util.extract_us(request)

        # Create a notification dispatch for each channel.
        for k, v in dispatch_mapping.items():
            notification_dispatch = self.notification_dispatch_cls(notification=notification,
                    due=due, category=k, view=v['view'], bcc=bcc,
                    single_spec=v['single'], batch_spec=v['batch'], address=email)
            session.add(notification_dispatch)

        # Save to the database.
        session.flush()

        return notification

class LookupNotificationDispatch(object):
    """Lookup notifications dispatch."""

    def __init__(self, **kwargs):
        self.model_cls = kwargs.get('model_cls', orm.NotificationDispatch)

    def __call__(self, id_):
        """Lookup by notifiction dispatch id."""

        return self.model_cls.query.get(id_)

    def by_notification_id(self, id_, type=u'email'):
        """Lookup all notification dispatches that belong to
        the notification id and type."""

        return self.model_cls.query.filter_by(notification_id=id_).all()

def get_or_create_notification_preferences(user):
    """Gets or creates the notification preferences for the user."""
    notification_preference_factory = NotificationPreferencesFactory()
    preference = user.notification_preference
    if preference is None:
        preference = notification_preference_factory(user.id)
        bm.Session.add(user)
    return preference

class NotificationPreferencesFactory(object):
    """Boilerplate to create and save ``Notification preference``s."""

    def __init__(self, **kwargs):
        self.notification_preference_cls = kwargs.get('notification_preference_cls',
                orm.NotificationPreference)
        self.session = kwargs.get('session', bm.Session)

    def __call__(self, user_id, frequency=None, channel='email'):
        """Create and store a notification and a notification dispatch."""

        # Unpack.
        session = self.session

        # Create notification.
        notification_preference = self.notification_preference_cls(
                user_id=user_id, frequency=frequency, channel=channel)

        # Save to the database.
        session.add(notification_preference)
        session.flush()

        return notification_preference



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name = 'pyramid_torque_engine_notifications',
    version = '0.0.1',
    description = 'Pyramid torque engine notification system based on nTorque based dual queue work engine system.',
    author = 'Andre Prado',
    author_email = 'username: andreprado88, domain: gmail.com',
    url = 'http://github.com/andrecp/pyramid_torque_engine_notifications',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Framework :: Pylons',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
    ],
    packages = find_packages('src'),
    package_dir = {'': 'src'},
    include_package_data = True,
    zip_safe = False,
    install_requires=[
        'fysom',
        'pyramid_basemodel',
        'pyramid_simpleauth',
        'pyramid_torque_engine',
        'pyramid_postmark',
        'transaction',
        'zope.interface'
    ],
    tests_require = [
        'coverage',
        'nose',
        'mock'
    ],
    entry_points = {
        'console_scripts': [
            'pyramid_notification = pyramid_torque_engine_notifications.notification_executer:run'
        ]
    }
)



# -*- coding: utf-8 -*-

"""
"""

__all__ = [
    'Notification',
    'NotificationDispatch',
    'NotificationPreference',
]

import os

from datetime import datetime

from sqlalchemy import event
from sqlalchemy import orm
from sqlalchemy import schema
from sqlalchemy import sql
from sqlalchemy import types
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext import associationproxy as proxy
from sqlalchemy.ext import declarative
from sqlalchemy.ext import hybrid

import pyramid_basemodel as bm
from pyramid_simpleauth import model as simpleauth_model

# XXX make settings configurable.
from pyramid_torque_engine.orm import ActivityEvent


class NotificationDispatch(bm.Base, bm.BaseMixin):
    """A notification dispatch to an user, holds information about how to deliver
    and when."""

    __tablename__ = 'notifications_dispatch'

    # Has a due date.
    due = schema.Column(types.DateTime)

    # Has a sent date.
    sent = schema.Column(types.DateTime)

    # has a Notification.
    notification_id = schema.Column(
        types.Integer,
        schema.ForeignKey('notifications.id'),
    )

    # bcc info
    bcc = schema.Column(types.Unicode(96))
    # view  -> function to decode things
    view = schema.Column(types.Unicode(96))
    # simple for the moment, either single or batch text. XXX use ENUM.
    type_ = schema.Column(types.Unicode(96))
    # dotted path for the asset spec.
    single_spec = schema.Column(types.Unicode(96))
    batch_spec = schema.Column(types.Unicode(96))
    # simple for the moment, either email or sms. XXX use ENUM.
    category = schema.Column(types.Unicode(96))
    # email or telephone number
    address = schema.Column(types.Unicode(96))

class Notification(bm.Base, bm.BaseMixin):
    """A notification about an event that should be sent to an user."""

    __tablename__ = 'notifications'

    # has an user.
    user_id = schema.Column(
        types.Integer,
        schema.ForeignKey('auth_users.id'),
    )

    user = orm.relationship(
        simpleauth_model.User,
        backref='notification',
    )

    # Has a read date.
    read = schema.Column(types.DateTime)

    notification_dispatch = orm.relationship(
        NotificationDispatch,
        backref='notification')

    # has an Activity event.
    # One to many
    event_id = schema.Column(
        types.Integer,
        schema.ForeignKey('activity_events.id'),
    )
    event = orm.relationship(
        ActivityEvent,
        backref=orm.backref(
            'notification',
        ),
    )

    def __json__(self, request=None):
        """Represent the event as a JSON serialisable dict."""

        data = {
            'id': self.id,
            'user_id': self.user_id,
            'created_at': self.created.isoformat(),
            'read_at': self.read.isoformat(),
            'event_id': self.event_id,
        }
        return data

class NotificationPreference(bm.Base, bm.BaseMixin):
    """Encapsulate user's notification preferences."""

    __tablename__ = 'notification_preferences'

    # Belongs to a user.
    user_id = schema.Column(types.Integer, schema.ForeignKey('auth_users.id'))
    user = orm.relationship(simpleauth_model.User, single_parent=True,
            backref=orm.backref('notification_preference', single_parent=True, uselist=False))

    # Optional notification preferences.
    # simple for the moment, either sms or email text. XXX use ENUM.
    channel = schema.Column(types.Unicode(96))
    # simple for the moment, either daily or weekly. XXX use ENUM.
    frequency = schema.Column(types.Unicode(96))

    def __json__(self, request=None):
        return {
            'id': self.id,
            'frequency': self.frequency,
            'channel': self.channel,
            'user_id': self.user_id,
        }



# -*- coding: utf-8 -*-

"""Provide a boilerplate factory to make the WSGI app to be tested."""

__all__ = [
    'TestAppFactory',
    'make_wsgi_app',
]

import unittest

try:
    import webtest
except ImportError:  # pragma: no cover
    pass

import pyramid_basemodel as bm

import transaction

from cornice.tests import support
from sqlalchemy import engine

from pyramid import config as pyramid_config
from pyramid import registry
from pyramid import testing

from pyramid_torque_engine import action
from pyramid_torque_engine import client
from pyramid_torque_engine import traverse

from pyramid_torque_engine import repo
from pyramid_simpleauth import model as simpleauth_model

from . import settings

# We monkey patch user just because..
class Email(object):
    address = 'testing@test.com'

simpleauth_model.User.best_email = Email()

def make_wsgi_app(root_factory, includeme, registry=None, **settings):
    """Create and return a WSGI application."""

    configurator_cls = pyramid_config.Configurator

    # Initialise a ``Configurator`` and apply the package configuration.
    if registry:
        config = configurator_cls(registry=registry)
        config.setup_registry(settings=settings, root_factory=root_factory)
    else:
        config = configurator_cls(settings=settings, root_factory=root_factory)

    # Include the app config.
    includeme(config)

    # Close any db connection for this thread.
    bm.Session.remove()

    # Return a WSGI app.
    return config.make_wsgi_app()

# Authentication related
def createUser(**kwargs):
    name = kwargs.get('name', u'Dummy')
    password = kwargs.get('password', u'password')
    data = {
            'password': password,
            'username': name,
        }
    user = simpleauth_model.User(**data)
    bm.Session.add(user)
    bm.Session.flush()
    return user

def createEvent(context=None):
    """Stubs an event."""
    from pyramid_torque_engine.repo import ActivityEventFactory
    from mock import Mock

    with transaction.manager:
        bm.Session.add(context)
        event = ActivityEventFactory(Mock())(context, None)
        bm.Session.add(event)
        bm.Session.flush()
        event_id = event.id
    return event_id

class StubRequest(object):
    """Provide `request.registry` and `request.environ`."""

    def __init__(self, registry_):
        self.registry = registry_
        self.environ = {'paste.testing': True,}
        self.torque = client.get_torque_api(self)

    @property
    def state_changer(self):
        return action.StateChanger(self)

    def get_state_machine(self, *args, **kwargs):
        return action.get_state_machine(self, *args, **kwargs)

    def __repr__(self):
        return 'StubRequest {0} {1}'.format(self.registry, self.environ)

class TestAppFactory(object):
    """Callable utility that returns a testable WSGI app and manages db state."""

    def __init__(self, app_factory, should_use_tx=True, **kwargs):
        self.app_factory = app_factory
        self.should_use_tx = should_use_tx
        self.base = kwargs.get('base', bm.Base)
        self.catch_errors = kwargs.get('catch_errors', support.CatchErrors)
        self.global_config = kwargs.get('global_config', None)
        self.json_method = kwargs.get('get_json', webtest.utils.json_method)
        self.session = kwargs.get('session', bm.Session)
        self.test_app = kwargs.get('test_app', webtest.TestApp)
        self.test_settings = kwargs.get('test_settings', settings.TEST_SETTINGS)
        self.has_created = False
        self.engine = engine.engine_from_config(self.test_settings,
                prefix='sqlalchemy.')
        self.conn = self.engine.connect()

    def begin(self):
        if self.should_use_tx:
            self.tx = self.conn.begin()
        self.session.configure(bind=self.conn)
        reg = registry.Registry(name='ftesting')
        self.configurator = testing.setUp(registry=reg)

    def rollback(self):
        self.session.remove()
        if self.should_use_tx:
            self.tx.rollback()
        testing.tearDown()

    def __call__(self, **kwargs):
        """Create the WSGI app and wrap it with a patched webtest.TestApp."""

        # Patch TestApp.
        self.test_app.get_json = self.json_method('GET')
        self.test_app.put_json = self.json_method('PUT')
        self.test_app.delete_json = self.json_method('DELETE')

        # Instantiate.
        self.settings = self.test_settings.copy()
        self.settings.update(kwargs)
        factory = self.app_factory()
        reg = self.configurator.registry
        app = factory(self.global_config, registry=reg, **self.settings)

        # Wrap.
        test_app = self.test_app(self.catch_errors(app))

        # Patch into the registry.
        self.configurator.registry.settings['webtest_app'] = test_app

        # Return the wrapped WSGI app.
        test_app.registry = reg
        return test_app

class AppTestCase(unittest.TestCase):
    """Unit test case that sets up a work engine WSGI app configured
      using ``cls.includeme(config)``.
    """

    @classmethod
    def includeall(cls, config):
        """Setup the db and work engine and include the app config."""

        # Boilerplate.
        config.include('pyramid_basemodel')
        config.include('pyramid_tm')
        config.include('pyramid_torque_engine')

        # App specifics.
        cls.includeme(config)

    @classmethod
    def includeme(cls, config):
        raise NotImplementedError

    @classmethod
    def setup_class(cls):
        class AppFactory(object):
            def __call__(self, global_config, **settings):
                return make_wsgi_app(traverse.EngineRoot, cls.includeall, **settings)
        cls.factory = TestAppFactory(AppFactory)

    def setUp(self):
        self.factory.begin()

    def tearDown(self):
        self.factory.rollback()

    def getRequest(self, app):
        """Return a stubbed request that has the configured registry and
          provides a minimal api.
        """

        return StubRequest(app.registry)
